##############################################################################
#
# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""$Id: OODocument.py,v 1.10 2003/02/08 22:19:16 longsleep Exp $
"""

# html rendered OpenOffice Documents for CMF/Plone by Simon Eisenmann

import Globals, string, re, os
from Globals import DTMLFile, InitializeClass
from StringIO import StringIO
from AccessControl import ClassSecurityInfo, getSecurityManager

import Acquisition
from Acquisition import aq_base
from OFS.Traversable import Traversable
from webdav.Collection import Collection
from ExtensionClass import Base
from OFS.content_types import guess_content_type

from Products.CMFCore.PortalContent import NoWL, ResourceLockedError
from Products.CMFCore.CMFCorePermissions import View
from Products.CMFCore.CMFCorePermissions import ModifyPortalContent
from Products.CMFCore.WorkflowCore import WorkflowAction
from Products.CMFCore.utils import keywordsplitter, _verifyActionPermissions, getToolByName
from Products.CMFDefault import utils
from Products.CMFDefault.Document import Document
from Products.CMFDefault.File import File
from Products.CMFDefault.DublinCore import DefaultDublinCoreImpl
from Products.CMFDefault.utils import parseHeadersBody, formatRFC822Headers
from Products.CMFDefault.utils import SimpleHTMLParser, bodyfinder, _dtmldir
from DocumentTemplate.DT_Util import html_quote

import OFS.Image

import libxml2
import libxslt
import zipfile
import tempfile

_DEBUG_XSLT = 0
text_format = 'application/vnd.sun.xml.writer'

# deactivate libxml2 and libxslt error messages
if not _DEBUG_XSLT:
    def noerr(ctx, str):
        pass
    libxml2.registerErrorHandler(noerr, None)
    libxslt.registerErrorHandler(noerr, None)

# load the openoffice 2 html transformation stylesheet
packageDir=Globals.package_home(globals())
__stylefile=packageDir.split(os.sep)+['sx2ml', 'main_html.xsl']
__stylefile='file:///'+'/'.join(__stylefile)
__styledoc=libxml2.parseFile(__stylefile)
_style=libxslt.parseStylesheetDoc(__styledoc)


class DynZipWrapper(Base):

    def __of__(self, parent):
        return parent.document()


class DynZip(Acquisition.Implicit, Traversable):
	""" temporary object to access the zipped file contents
            like images on demand
	"""
	def __init__(self, zip, current=''):
		self.zip=zip
		self.current=current
		
        def read(self, name):
	        return self.zip.read(name)

	def close(self):
		return self.zip.close()

	def __getitem__(self, name):
		if self.current: spacer='/'
		else: spacer=''
		name='%s%s%s' % (self.current, spacer, name)
		return DynZip(self.zip, current=name).__of__(self)

	def __call__(self, *args, **kw):
		"""
		"""
		REQUEST=self.REQUEST
		body=self.read(self.current)
		fn=self.current.split('/')[-1]
		content_type, enc=guess_content_type(fn, body)
		REQUEST.RESPONSE.setHeader('Content-Type', content_type)
	        REQUEST.RESPONSE.setHeader('Content-Disposition', 'filename=%s' % (fn))
		try: return self.read(self.current)
		except: raise

	index_html=__call__


factory_type_information = ( { 'id'             : 'OODocument'
                             , 'meta_type'      : 'OODocument'
                             , 'description'    : """\
OODocuments contain Documents written with OpenOffice Writer. The content is rendered as html."""
                             , 'icon'           : 'document_icon.gif'
                             , 'product'        : 'CMFOODocument'
                             , 'factory'        : 'addOODocument'
                             , 'immediate_view' : 'metadata_edit_form'
                             , 'actions'        :
                                ( { 'id'            : 'view' 
                                  , 'name'          : 'View'
                                  , 'action'        : 'document_view'
                                  , 'permissions'   : (View,)
                                  }
 				, { 'id'	    : 'download'
	                          , 'name'	    : 'Download'
				  , 'action'	    : 'file_view'
				  , 'permissions'   : (View,)
				  , 'visible'	    : 0
				  }
                                , { 'id'            : 'edit'
                                  , 'name'          : 'Edit'
                                  , 'action'        : 'file_edit_form'
                                  , 'permissions'   : (ModifyPortalContent,)
                                  }
                                , { 'id'            : 'metadata'
                                  , 'name'          : 'Metadata'
                                  , 'action'        : 'metadata_edit_form'
                                  , 'permissions'   : (ModifyPortalContent,)
                                  }
                                , { 'id'            : 'external_edit'
                                  , 'name'          : 'External Edit'
                                  , 'action'        : 'external_edit'
                                  , 'permissions'   : (ModifyPortalContent,)
				  , 'visible'	    : 0
                                  }
                                )
                             }
                           ,
                           )

def addOODocument( self
                 , id
		 , title=''
		 , file=''
		 , content_type=''
		 , precondition=''
		 , subject=() 
		 , description=''
		 , contributors=()
		 , effective_date=None
		 , expiration_date=None
		 , format=text_format
		 , language=''
		 , rights=''
                 ):
    """ Add a OODocument """

    # cookId sets the id and tile if they are not set
    id, title = OFS.Image.cookId(id, title, file)

    self=self.this()

    o = OODocument( id, title, '', content_type, precondition, subject
                   , description, contributors, effective_date, expiration_date
                   , format, language, rights
                   )
   
    # Add the OODocument instance to self
    self._setObject(id, o)

    # 'Upload' the file. This is done now rather than in the
    # contructor because the object is now in the ZODB and
    # can span ZODB objecs
    self._getOb(id).manage_upload(file)


class OODocument(File):
    """ A Document which holds and renders OpenOffice files to html """

    meta_type = 'OODocument'
    effective_date = expiration_date = None
    _isDiscussable = 1
    text='true' # this is set for compatibility with document_view
    text_format='html'

    security = ClassSecurityInfo()

    security.declareProtected(View, 'index_html')
    def index_html(self, REQUEST=None, RESPONSE={}):
	"""
	Invokes the default view.
	"""
	view = _getViewFor(self, 'view', 'document_view')
	if getattr(aq_base(view), 'isDocTemp', 0):
	    return apply(view, (self, self.REQUEST))
	else:
	    return view()

    __call__=view=index_html

    security.declareProtected(View, 'CookedBody')
    def CookedBody(self, stx_level=None, setlevel=0):
	"""
	return openoffice file as html
	"""
	return getattr(self, 'cooked_text', 'This OpenOffice Document has yet to be uploaded.')

    security.declareProtected(View, 'Description')
    def Description(self):
	"""
	Dublin core description
	"""
	return self.description

    security.declareProtected(View, 'SearchableText')
    def SearchableText(self):
        """ Used by the catalog for basic full text indexing """
        return "%s %s %s" % ( self.Title()
                            , self.Description()
                            , self.CookedBody()
                            )
    security.declareProtected(ModifyPortalContent, 'edit')
    def edit(self, precondition='', file=''):
	""" Update and reindex. """
	self._edit( precondition, file)
	self.reindexObject()

    security.declarePrivate('_edit')
    def _edit(self, precondition='', file=''):
        """ Edit the OODocument - Parses headers and cooks the body """

        if precondition: self.precondition = precondition
        elif self.precondition: del self.precondition

        if self._isNotEmpty(file):
            self.manage_upload(file)

        self.setFormat(text_format)

	# cook it
	self._cook()

    security.declarePrivate('_cook')
    def _cook(self):
	""" render the document by xslt to html """

	params={}
	# default params
	baseURL='doc'
	params.update({  'absoluteSourceDirRef': "'%s'" % (baseURL)
		       , 'disableJava': 'true()'
		       , 'disableLinkedTableOfContent': 'true()'
		     })

	if _DEBUG_XSLT:
	    params.update({'DEBUG': 'true()'})

	# get the zipped document
	document=self.document()

	# create temp files
	tmpfiles=[]
	for f in (('stylesFileURL','styles.xml'), ('metaFileURL','meta.xml')):
	    # get unique filename and write contents to this file
            t=tempfile.mktemp()
            tmpfiles.append(t)
            file=open(t, 'wb', -1)
            file.write(document.read(f[1]))
            file.close()
	    # update the parameters to reflect the temporary files
            params.update({f[0]: "'%s'" % (t)})

	# read the content xml
	thisdoc=document.read('content.xml')
	# parse the content
	doc=libxml2.parseDoc(thisdoc)
	# apply the stylesheet (xslt transformation)
	result=_style.applyStylesheet(doc, params)
	# write html to file
	html=result.serialize()
	# free memory
	result.freeDoc()
	doc.freeDoc()
	# write rendered document to zodb
	self._storeCooked(body=html)
	# clean up created temp files
	for t in tmpfiles:
	    try: os.unlink(t)
	    except: pass

    security.declarePrivate('_storeCooked')
    def _storeCooked(self, body=None):
	""" parse the rendered document and store metadata and body """
	if body == None: body=self.CookedBody()
	headers, body = self.handleText(text=body)
	#print "headers", headers
	self.cooked_text=body
	self.setMetadata(headers)

    security.declarePrivate('handleText')
    def handleText(self, text):
	""" Handles the raw html content returning headers and body """
	headers = {}
	# create html parser
	parser = SimpleHTMLParser()
	# parse for headers and body
	parser.feed(text)
	# update headers dictionary
	headers.update(parser.metatags)
	if parser.title:
	    headers['Title'] = parser.title
	if not getattr(self, 'keepOOStyles', 0):
	    # remove html header (including oo styles)
            bodyfound = bodyfinder(text)
	    if bodyfound: text=bodyfound
	return headers, text

    security.declarePrivate('update_data')
    def update_data(self, data, content_type=None, size=None):
    	# print "content_type", content_type
        if content_type is not None: self.content_type=text_format
        if size is None: size=len(data)
        self.size=size
        self.data=data
        self.ZCacheable_invalidate()
        self.http__refreshEtag()

    security.declareProtected(View, 'document')
    def document(self):
	"""
	returns a dynzip instance
	"""
	# XXX: how is that zipfile and stringio instance closed?
	zip=zipfile.ZipFile(StringIO(self.data), 'rb') 
	return DynZip(zip).__of__(self)

    # make zip contents accessable via web
    doc=DynZipWrapper()

    security.declareProtected(ModifyPortalContent, 'setMetadata')
    setMetadata = Document.setMetadata

    security.declareProtected(ModifyPortalContent, 'PUT')
    def PUT(self, REQUEST, RESPONSE):
        """ Handle HTTP PUT requests """
        self.dav__init(REQUEST, RESPONSE)
        self.dav__simpleifhandler(REQUEST, RESPONSE, refresh=1)
        type=REQUEST.get_header('content-type', None)

        file=REQUEST['BODYFILE']

        data, size = self._read_data(file)
        content_type=self._get_content_type(file, data, self.__name__,
                                            type or self.content_type)
        self.update_data(data, content_type, size)
	self._cook()

        RESPONSE.setStatus(204)
        return RESPONSE



def _getViewFor(obj, view='view', default=None):
    ti = obj.getTypeInfo()
    if ti is not None:
        actions = ti.getActions()
        for action in actions:
            if action.get('id', None) == default:
                default=action
            if action.get('id', None) == view:
                if _verifyActionPermissions(obj, action) and action['action']!='':
                    return obj.restrictedTraverse(action['action'])

        if default is not None:
            if _verifyActionPermissions(obj, default):
                return obj.restrictedTraverse(default['action'])

        raise 'Unauthorized', ('No accessible views available for %s' %
                               '/'.join(obj.getPhysicalPath()))
    else:
        raise 'Not Found', ('Cannot find default view for "%s"' %
                            '/'.join(obj.getPhysicalPath()))

InitializeClass(OODocument)
